'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var redux = require('redux');
var produce = require('immer');
var produce__default = _interopDefault(produce);

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectSpread(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};
    var ownKeys = Object.keys(source);

    if (typeof Object.getOwnPropertySymbols === 'function') {
      ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) {
        return Object.getOwnPropertyDescriptor(source, sym).enumerable;
      }));
    }

    ownKeys.forEach(function (key) {
      _defineProperty(target, key, source[key]);
    });
  }

  return target;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  }
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArray(iter) {
  if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
}

function _iterableToArrayLimit(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance");
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

var get = function get(path, target) {
  return path.reduce(function (acc, cur) {
    return acc[cur];
  }, target);
};

var isPromise = function isPromise(x) {
  return _typeof(x) === 'object' && typeof x.then === 'function';
};

var hasNestedAction = function hasNestedAction(current) {
  return Object.keys(current).reduce(function (acc, key) {
    var value = current[key];

    if (typeof value === 'function') {
      return true;
    }

    if (_typeof(value) === 'object' && !Array.isArray(value)) {
      return acc || hasNestedAction(value);
    }

    return acc;
  }, false);
};

var extractInitialState = function extractInitialState(current) {
  return Object.keys(current).reduce(function (innerAcc, key) {
    var value = current[key];

    if (typeof value === 'function') {
      return innerAcc;
    }

    return _objectSpread({}, innerAcc, _defineProperty({}, key, _typeof(value) === 'object' && !Array.isArray(value) ? extractInitialState(value) : value));
  }, {});
};

var easyPeasy = function easyPeasy(model) {
  var definition = _objectSpread({}, model, {
    logFullState: function logFullState(state) {
      // eslint-disable-next-line no-console
      console.log(JSON.stringify(state, null, 2));
    }
  });

  var references = {};

  var extractActionHandlers = function extractActionHandlers(current, path) {
    return Object.keys(current).reduce(function (innerAcc, key) {
      var value = current[key];

      if (typeof value === 'function') {
        var handler = function handler(state, payload) {
          var inverseChanges = [];
          var inverse = false;
          var newState = produce__default(state, function (draft) {
            var handlerResult = value(draft, payload, {
              dispatch: references.dispatch,
              dispatchLocal: get(path, references.dispatch),
              getState: references.getState
            });
            inverse = isPromise(handlerResult);
          }, function (_, inversePatches) {
            inverseChanges.push.apply(inverseChanges, _toConsumableArray(inversePatches));
          });
          return inverse ? produce.applyPatches(newState, inverseChanges) : newState;
        };

        handler.actionName = "".concat(path.join('.'), ".").concat(key);
        return _objectSpread({}, innerAcc, _defineProperty({}, key, handler));
      }

      if (_typeof(value) === 'object' && !Array.isArray(value)) {
        var _actions = extractActionHandlers(value, _toConsumableArray(path).concat([key]));

        if (hasNestedAction(_actions)) {
          return _objectSpread({}, innerAcc, _defineProperty({}, key, _actions));
        }
      }

      return innerAcc;
    }, {});
  };

  var extractActions = function extractActions(current) {
    return Object.keys(current).reduce(function (innerAcc, key) {
      var value = current[key];

      if (typeof value === 'function') {
        var action = function action(payload) {
          return references.dispatch({
            type: value.actionName,
            payload: payload
          });
        };

        return _objectSpread({}, innerAcc, _defineProperty({}, key, action));
      }

      if (_typeof(value) === 'object' && !Array.isArray(value)) {
        return _objectSpread({}, innerAcc, _defineProperty({}, key, extractActions(value)));
      }

      return innerAcc;
    }, {});
  };

  var initialState = extractInitialState(definition);
  var actionHandlers = extractActionHandlers(definition, []);
  var actions = extractActions(actionHandlers);

  var extractReducer = function extractReducer(current, path) {
    var handlersAtPath = Object.keys(current).reduce(function (acc, key) {
      var value = current[key];

      if (typeof value === 'function') {
        return _toConsumableArray(acc).concat([value]);
      }

      return acc;
    }, []);
    var stateAtPath = Object.keys(current).reduce(function (acc, key) {
      var value = current[key];

      if (_typeof(value) === 'object' && !Array.isArray(value)) {
        return _toConsumableArray(acc).concat([key]);
      }

      return acc;
    }, []);

    if (handlersAtPath.length > 0) {
      var nestedReducers = stateAtPath.map(function (key) {
        return [key, extractReducer(current[key], _toConsumableArray(path).concat([key]))];
      });
      var defaultState = get(path, initialState);
      return function () {
        var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : defaultState;
        var action = arguments.length > 1 ? arguments[1] : undefined;
        var handler = handlersAtPath.find(function (x) {
          return x.actionName === action.type;
        });

        if (handler) {
          return handler(state, action.payload);
        }

        for (var i = 0; i < nestedReducers.length; i += 1) {
          var _nestedReducers$i = _slicedToArray(nestedReducers[i], 2),
              key = _nestedReducers$i[0],
              reducer = _nestedReducers$i[1];

          var newState = reducer(state[key], action);

          if (state[key] !== newState) {
            return _objectSpread({}, state, _defineProperty({}, key, newState));
          }
        }

        return state;
      };
    }

    return redux.combineReducers(stateAtPath.reduce(function (acc, key) {
      return _objectSpread({}, acc, _defineProperty({}, key, extractReducer(current[key], _toConsumableArray(path).concat([key]))));
    }, {}));
  };

  var store = redux.createStore(extractReducer(actionHandlers, []), initialState); // attach the actions to dispatch

  Object.keys(actions).forEach(function (key) {
    store.dispatch[key] = actions[key];
  });
  references.dispatch = store.dispatch;
  references.getState = store.getState;
  return store;
};

module.exports = easyPeasy;
//# sourceMappingURL=easy-peasy.js.map
